exception Bad_char of char
exception Bad_string of string
exception Bad_input
exception No_command
exception Pattern_lacks_origin

(****** Random useful stuff ******)

let hash_iter hash f = Hashtbl.iter f hash
let hash_values hash =
  let values = ref [] in
  hash_iter hash (fun k v -> values := v :: !values);
  !values
let hash_keys hash =
  let keys = ref [] in
  hash_iter hash (fun k v -> keys := v :: !keys);
  !keys

let list_exists list f = List.exists f list
let list_filter list f = List.filter f list
let list_find list f = List.find f list
let list_for_all list f = List.for_all f list
let list_iter list f = List.iter f list
let list_map list f = List.map f list
let list_sort list f = List.sort f list

let list_empty list =
  try
    ignore (List.hd list);
    false
  with
    | Failure "hd" -> true
let list_nonempty list  = not (list_empty list)

let rec iter_abort lst func =
  match lst with
  | [] -> None
  | a::b -> let res = func a in
       if res != None then res else iter_abort b func
;;

let sub_fold_left f x a start finish =
  let r = ref x and
      s = (if start < 0 then 0 else start) and
      e = (if finish > (Array.length a - 1) then (Array.length a - 1) else finish) in
  for i = s to e do
    r := f !r (Array.unsafe_get a i)
  done;
  !r
;;

(****** Main types ******)

(* Elements of the map *)
type square_type = SPACE | WALL | WATER | BASE

type square = {
  kind: square_type;
  is_beach: bool;
  mutable has_robots: bool;
}

type robot = {
  id: string;
  name: char;
  mutable capacity: int;
  mutable bankroll: int;
  mutable x: int;
  mutable y: int;
  mutable packages: int list;
  mutable history: (int*int) list;

  mutable bidwar: int;

  mutable alive: bool
}

type package_location =
    Sitting of int*int           (* ready to be picked up *)
  | Carried of robot             (* carried around by someone *)

type package = {
  pkg_id: int;
  mutable pkg_loc: package_location;    (* current location *)
  mutable pkg_dest: (int*int) option;   (* destination if known *)
  mutable weight: int option;           (* weight if known *)
}

type direction = N | E | S | W   (* A direction a robot can move in *)

type order =                     (* The actual order part of a command *)
    Move of direction
  | Pick of int list
  | Drop of int list
  | No_order

type command = {                 (* Order + bid *)
  bid: int;
  order: order
}

type update_action =             (* What happened to a robot *)
    UpdateMove of direction
  | UpdatePick of int
  | UpdateDrop of int
  | UpdateX of int
  | UpdateY of int

type update = {                  (* update_action along with who it happened to *)
  robot: robot;
  action: update_action
}

type matcher =
    IGNORE
  | PLAYER
  | MAYBE_BEACH
  | OPEN_NOT_BEACH
  | ROBOT_N
  | ROBOT_E
  | ROBOT_S
  | ROBOT_W
  | ROBOT_STILL   (* robot MIGHT not move *)
  | ROBOT_ROCK    (* robot ALMOST CERTAINLY will not move *)
  | IS_WALL_OR_WATER
  | IS_WATER
  | ISNT_WATER

type pattern = {
   pat_action: order;
   pat: matcher array array;
   mutable source: int*int;
}

type io_interface = {            (* socket and log file *)
  sock: Unix.file_descr;
  log:  out_channel
}

(****** Operations on types ******)

let is_walkable square =
  match square with | BASE -> true | SPACE -> true | _ -> false
;;

let is_water square = (square == WATER) ;;

(* Square to printed representation and back again *)
let square_ch = function SPACE -> '.' | WATER -> '~' | WALL -> '#' | BASE -> '@'
let square_ch_beach = function SPACE -> ':' | WATER -> '~' | WALL -> '#' | BASE -> '0'
let ch_square = function '.' -> SPACE | '~' -> WATER | '#' -> WALL | '@' -> BASE
  | c -> raise (Bad_char c)

(* Printed representation of directions *)
let dir_to_str = function N -> "N" | E -> "E" | S -> "S" | W -> "W"

let dir_to_deltas = function N -> (0,1) | E -> (1,0) | S -> (0,-1) | W -> (-1,0)

let deltas_to_dir = function (0,1) -> N | (1,0) -> E | (0,-1) -> S | (-1,0) -> W
  | _ -> raise Not_found

(* Printed representation of orders *)
let order_to_str = function
  | Move( dir ) -> "Move " ^ (dir_to_str dir)
  | Pick( lst ) -> "Pick " ^ String.concat " " (list_map lst string_of_int)
  | Drop( lst ) -> "Drop " ^ String.concat " " (list_map lst string_of_int)
  | No_order -> "No Move [BUG]"

(* Printed representation of commands *)
let command_to_str cmd = (string_of_int cmd.bid) ^ " " ^ (order_to_str cmd.order)

(* makes a blank width*height board *)
let make_board width height =
  Array.init (width+2) (fun i -> Array.make (height+2) {kind=SPACE;is_beach=false;has_robots=false})

let new_robot id = 
  { id=id;
    name=String.get "abcdefghijklmnopqrstuvwxyz" ((int_of_string id) mod 26);
    capacity=0; bankroll=0;
    x=0; y=0;
    packages=[];
    bidwar=2;
    history=[];
    alive=true;
  }

let package_weight p = match p.weight with Some(w) -> w | None -> 1
let package_dest p = match p.pkg_dest with Some(x,y) -> (x,y) | None -> (-1,-1)

(* We just found out that robot is carrying the package with id p *)
let get_or_make_package p packages loc =
  try Hashtbl.find packages p
  with Not_found ->
    let pack = {pkg_id=p; pkg_loc=loc; pkg_dest=None; weight=None} in
    Hashtbl.add packages p pack;
    pack

let package_location_to_str = function
    Sitting(x,y) -> "[" ^ (string_of_int x) ^ "," ^ (string_of_int y) ^ "]"
  | Carried(r) -> "{" ^ (r.id) ^ "}"

let package_dest_to_str = function
    Some(x,y) -> "[" ^ (string_of_int x) ^ "," ^ (string_of_int y) ^ "]"
  | None -> "?"

let package_weight_to_str = function
    Some(w) -> (string_of_int w)
  | None -> "?"

let package_to_str p =
  (string_of_int p.pkg_id) ^ ": " ^
  (package_location_to_str p.pkg_loc) ^ " -> " ^
  (package_dest_to_str p.pkg_dest) ^ " (" ^
  (package_weight_to_str p.weight) ^ ")"


