open Dib_types
open Dib_io
open Sock_io

(*
  To do:

  HIGHEST PRIORITY
  - Move if in danger of being dunked, instead of drop/pickup
                                      (also instead of doomed-to-fail-move)
    (I try to avoid being dunked a little bit with the patterns)

  OTHER
  - In pathfinding, try to go through interesting squares if possible
  - it would be good to test if a guy is in the place we're going, and 
    may_go says he's not moving towards us, then we prefer to bid negative
    (any other predictive things in bidding?)
  - don't do this kind of dunking because it's unsafe on the next turn
    (if the other robot advances next to you, you're on the beach)
     RrR~
    so avoid any dunking when there is a robot next to you 
    (besides the dunkee)
  - bidwar bids go up as the game progresses, but near the end of the
    game they should come back down, e.g. continually clamp them to
    the max_bidwar computation I've included
*)

let sock_log = ref stdout;;

let max_dist = 2*(1000*1000+1)  ;;
let max_history = 4             ;;

(****** Reasoning about robots behavior ******)

(* list must not be empty! *)
let all_identical list =
  list_for_all list (fun x -> x = List.hd list) ;;

let for_all_pairwise the_list func =
  let result = ref true 
  and list = ref the_list in
  while !list <> [] do
    match !list with
      | a::b::rest ->
        (if not (func a b) then result := false);
        list := b::rest
      | a::rest -> list := []
      | _ -> ()
  done;
  !result

let is_rock r =
  if List.length r.history < max_history then false
  else let full = (r.x, r.y) :: r.history in
  (all_identical full)
;;
  

(* predict whether robot r may go to (x,y), assuming
   robot r is within reach of (x,y)
*)
let may_go r (x, y) =
  if List.length r.history < max_history then true
  else let full = (r.x, r.y) :: r.history in

  (* is he a rock? *)
  if all_identical full then
    if x == r.x && y == r.y then true
    else false

  else if x == r.x && y == r.y then begin

    (* did he stop for one turn to drop/pickup? *)
    (* not enough knowledge, so we fake this    *)
    let prev = List.nth r.history 1 in
    if (x,y) = prev then false

    (* otherwise, he's just on the move, and we have no clue *)
    else true
  end else begin
    (* look for a consistent trend in x and in y *)
    let (lx,ly) = List.nth full max_history in

    let x_mono_inc = for_all_pairwise full (fun (x1,y1) (x2,y2) -> (x1 >= x2))
    and x_mono_dec = for_all_pairwise full (fun (x1,y1) (x2,y2) -> (x1 <= x2))
    and y_mono_inc = for_all_pairwise full (fun (x1,y1) (x2,y2) -> (y1 >= y2))
    and y_mono_dec = for_all_pairwise full (fun (x1,y1) (x2,y2) -> (y1 <= y2)) in
    (* r.x,r.y is the newest *)
    (* lx,ly is the oldest *)
    (* robot is going right if r.x > lx and x_mono_inc *)

(* the old way requires monotonic increasing AND non-zero increase
         if r.x < x && lx > r.x && x_mono_dec then false
    else if r.x > x && lx < r.x && x_mono_inc then false
    else if r.y < y && ly > r.y && y_mono_dec then false
    else if r.y > y && ly < r.y && y_mono_inc then false
*)
         if r.x < x && x_mono_dec then false
    else if r.x > x && x_mono_inc then false
    else if r.y < y && y_mono_dec then false
    else if r.y > y && y_mono_inc then false
    else
      true
  end
;;

let predict_dir out r dir =
  let (dx,dy) = dir_to_deltas dir in
  let (x,y) = (r.x+dx, r.y+dy) in
  if may_go r (x, y) then output_string out ((dir_to_str dir)^" ")
;;

let make_predictions outp robots = hash_iter robots
  (fun k r ->
    Printf.fprintf outp "Predict %s: " r.id;
    predict_dir outp r N;
    predict_dir outp r E;
    predict_dir outp r S;
    predict_dir outp r W;
    if may_go r (r.x,r.y) then output_string outp "x ";
    output_char outp '\n';
    flush outp
  )
;;

(****** Pathfinding *******)

let make_inf_dists width height =
  Array.init (width) (fun i -> Array.make (height) max_dist) ;;

let print_dists dists =
  let (width, height) = (Array.length dists, Array.length (dists.(1))) in
  for j = 1 to height-2 do
    for i = 1 to width-2 do
      if dists.(i).(j) == max_dist then
        Printf.printf "- "
      else
        Printf.printf "%d " dists.(i).(j)
    done;
    Printf.printf "\n";
  done
;;

(* explore a space next to current space; if it's not been visited *)
(* yet, update the distance to it, and add it to the queue for visiting *)
let try_dir rule board dists queue dist x y =
  if dists.(x).(y) == max_dist then
    if rule board.(x).(y) then
      match board.(x).(y).kind with
        | SPACE  -> dists.(x).(y) <- dist; Queue.add (x,y) queue
        | BASE   -> dists.(x).(y) <- dist; Queue.add (x,y) queue
        | WATER  -> ()
        | WALL   -> ()

let normal x = true
let avoid_robots x = (not (x.has_robots))
let avoid_all x = not (x.has_robots || x.is_beach);;

(* recurse through the queue of spaces to explore from *)
let rec find_distances_rec rule board dists queue =
  if Queue.is_empty queue then
    dists
  else
    let (x,y) = Queue.take queue in
    let dist = dists.(x).(y)+1 in
      try_dir rule board dists queue dist (x+1) y;
      try_dir rule board dists queue dist (x-1) y;
      try_dir rule board dists queue dist x (y+1);
      try_dir rule board dists queue dist x (y-1);
      find_distances_rec rule board dists queue

(* find distances to every point on the board from x,y *)
let find_distances mode board x y =
  let dists = make_inf_dists (Array.length board) (Array.length board.(1)) in
  let queue = Queue.create() in
    Queue.add (x,y) queue;
    dists.(x).(y) <- 0;
    find_distances_rec mode board dists queue

(* given a distance table, find a route from x,y to the 0-point
   store the route in queue of (x, y, direction) where (x,y)
   is the point to move FROM--so you can validate that you're
   looking at the right route info first
*)

let prefer sq =
  (not sq.has_robots) && (not sq.is_beach)
;;

let rec find_route_rec board dists x y queue =
  if dists.(x).(y) == max_dist then
    Queue.clear queue
  else if dists.(x).(y) == 0 then
    ()
  else
    let dist = dists.(x).(y)-1 in

    (* prioritize spaces without robots--this doesn't seem to work!!! *)
         if dists.(x-1).(y) == dist && prefer board.(x-1).(y) then
           (Queue.add (x,y,W) queue; find_route_rec board dists (x-1) y queue)
    else if dists.(x).(y-1) == dist && prefer board.(x).(y-1) then
           (Queue.add (x,y,S) queue; find_route_rec board dists x (y-1) queue)
    else if dists.(x+1).(y) == dist && prefer board.(x+1).(y) then
           (Queue.add (x,y,E) queue; find_route_rec board dists (x+1) y queue)
    else if dists.(x).(y+1) == dist && prefer board.(x).(y+1) then
           (Queue.add (x,y,N) queue; find_route_rec board dists x (y+1) queue)

    (* now accept spaces with robots, in a different order just because *)
    else if dists.(x).(y-1) == dist then
           (Queue.add (x,y,S) queue; find_route_rec board dists x (y-1) queue)
    else if dists.(x-1).(y) == dist then
           (Queue.add (x,y,W) queue; find_route_rec board dists (x-1) y queue)
    else if dists.(x).(y+1) == dist then
           (Queue.add (x,y,N) queue; find_route_rec board dists x (y+1) queue)
    else if dists.(x+1).(y) == dist then
           (Queue.add (x,y,E) queue; find_route_rec board dists (x+1) y queue)
    else
       Queue.clear queue

let find_route board dists (x,y) =
  let queue = Queue.create() in
  find_route_rec board dists x y queue;
  queue

let queue_add x y =
  Queue.add x y;y

let rec reverse_route queue =
  if Queue.is_empty queue then Queue.create()
  else
    let item = Queue.take queue in
    match item with
      | (x,y,N) -> queue_add (x,y+1,S) (reverse_route queue)
      | (x,y,S) -> queue_add (x,y-1,N) (reverse_route queue)
      | (x,y,W) -> queue_add (x-1,y,E) (reverse_route queue)
      | (x,y,E) -> queue_add (x+1,y,W) (reverse_route queue)

let print_route queue =
  Printf.printf "[Route: ";
  Queue.iter (fun (x,y,d) -> Printf.printf "(%d,%d,%s) - " x y (dir_to_str d)) queue;
  Printf.printf "dest ]\n";flush stdout

(* compute a pathfind from robot to goal, dodging
   nearby robots and avoiding beaches *)

let route_around_bots board dists robots r (x, y) =

  (* first see if there's a path from robot to goal that avoids bots *)

  let avoid_dists = find_distances avoid_robots board x y in
  let avoid_dist = avoid_dists.(r.x).(r.y) in
  if avoid_dist = max_dist then

    (* no path except through the bot, so just try it *)

    (reverse_route (find_route board dists (x,y)), "can't avoid bot")

  else

    (* there is a path around... it could be a long way around,
       but oh well *)

    let beach_dists = find_distances avoid_all board x y in
    if beach_dists.(r.x).(r.y) < avoid_dist + (avoid_dist/8) + 2 then
      (* avoiding beaches too isn't that much longer, so go for it *)

      (find_route board beach_dists (r.x, r.y), "avoid beach")

    else
      (find_route board avoid_dists (r.x, r.y), "avoid bot")
;;

let avoid_range = 3;;

let any_nearby robots r =
  Hashtbl.fold (fun k v res ->
      let d = abs(v.x-r.x) + abs(v.y-r.y) in
      if d > 0 && d <= avoid_range then true else res)
    robots
    false;;

(*** this is what we did when we didn't have navigation ***)

let find_good_route_old board dists robots r dest =
  if any_nearby robots r
    then route_around_bots board dists robots r dest
    else (reverse_route (find_route board dists dest), "no bots")

(*** now we rely on the pattern matcher ***) 
let find_good_route board dists robots r dest =
    (reverse_route (find_route board dists dest), "")

(****** Actual logic!!! ******)

(****** navigation pattern matching *****)

(** patterns are stored as arrays of ROWS **)
(**   in other words, uses pat.(j).(i)    **)

let show_pattern out x =
  let (a,b) = x.source in
  Printf.fprintf out "Pattern with origin @ %d %d\n" a b;
  for j = (Array.length x.pat)-1 downto 0 do
    let z = x.pat.(j) in
    for i = 0 to (Array.length z)-1 do
      output_char out (match z.(i) with
        | IGNORE -> ' '
        | PLAYER -> '*'
        | MAYBE_BEACH -> ':'
        | OPEN_NOT_BEACH -> '.'
        | ROBOT_N -> '^'
        | ROBOT_E -> '>'
        | ROBOT_S -> 'v'
        | ROBOT_W -> '<'
        | ROBOT_STILL -> 'x'
        | ROBOT_ROCK  -> 'I'
        | IS_WALL_OR_WATER -> '#'
        | IS_WATER -> '~'
        | ISNT_WATER -> 'w')
    done; 
    output_char out '\n'
  done
;;

let patterns = try load_patterns "patterns.dat" with
  Sys_error(x) -> (print_endline "Pattern file not found" ; [| |] )
;;


(* if our move was to the N, and our direction pattern said to *)
(* go N for this pattern assuming an E move, then we should    *)
(* end up going W *)
let map_rot_dir d e = 
  match d with
    | E -> e
    | N -> (match e with | E -> N | N -> W | W -> S | S -> E)
    | W -> (match e with | E -> W | N -> S | W -> E | S -> N)
    | S -> (match e with | E -> S | N -> E | W -> N | S -> W)
;;

let match_pattern board rob_tab p (x, y) d =
  let (px, py) = p.source in
  (* map pattern coordinates to world coordinates *)
  (* if the move is east, it's unrotated *)
  (* if the move is north, then (0,1) in pattern goes to (-1,0) in world *)
  let map (nx, ny) = match d with
    | E -> ( (nx-px) + x,  (ny-py) + y)
    | N -> (-(ny-py) + x,  (nx-px) + y)
    | W -> (-(nx-px) + x, -(ny-py) + y)
    | S -> ( (ny-py) + x, -(nx-px) + y) in
  let res = ref true in
  for j = 0 to (Array.length p.pat)-1 do
    for i = 0 to (Array.length p.pat.(j))-1 do
      let (x,y) = map (i,j) in
      if (x <= 0 || y <= 0 || x >= -1+Array.length board || y >= -1+Array.length board.(1))
      then begin match p.pat.(j).(i) with
        | IGNORE -> ()
        | IS_WALL_OR_WATER -> ()
        | ISNT_WATER -> ()
        | _ -> res := false
      end else if not (match p.pat.(j).(i) with
        | IGNORE -> true
        | PLAYER -> true
        | MAYBE_BEACH ->
(*          Printf.fprintf !sock_log "Checking %d,%d (%d,%d)\n" x y i j;  *)
            is_walkable board.(x).(y).kind && rob_tab.(x).(y)==[]
        | OPEN_NOT_BEACH -> is_walkable board.(x).(y).kind && not board.(x).(y).is_beach && rob_tab.(x).(y)==[]
        | ROBOT_N -> list_exists rob_tab.(x).(y) (fun z -> may_go z (map (i,j+1)))
        | ROBOT_E -> list_exists rob_tab.(x).(y) (fun z -> may_go z (map (i+1,j)))
        | ROBOT_S -> list_exists rob_tab.(x).(y) (fun z -> may_go z (map (i,j-1)))
        | ROBOT_W -> list_exists rob_tab.(x).(y) (fun z -> may_go z (map (i+1,j)))
        | ROBOT_STILL -> list_exists rob_tab.(x).(y) (fun z -> may_go z (x,y))
        | ROBOT_ROCK  -> list_exists rob_tab.(x).(y) (fun z -> is_rock z)
        | IS_WALL_OR_WATER -> not (is_walkable board.(x).(y).kind)
        | ISNT_WATER -> not (board.(x).(y).kind == WATER)
        | IS_WATER -> board.(x).(y).kind == WATER
      )
      then res := false
    done;
  done;
  !res
;;

let build_robot_table robots width height =
  let tab = Array.init (width) (fun i -> Array.make (height) []) in
  Hashtbl.iter (fun k r -> tab.(r.x).(r.y) <- r::tab.(r.x).(r.y)) robots;
  tab
;;

(* run through patterns backwards so we keep the lowest #ed match *)

let pattern_match_navigate board robots r d =
  let robot_table = build_robot_table robots (Array.length board) (Array.length board.(1)) in
  let default = Move(d) in
  let best_cmd = ref default in
  for x = (Array.length patterns)-1 downto 0 do
    if match_pattern board robot_table patterns.(x) (r.x, r.y) d then begin
      Printf.printf            "Matched pattern %d!\n" x;
(*      Printf.fprintf !sock_log "Matched pattern %d was %s!\n" x (dir_to_str d);
      show_pattern !sock_log patterns.(x); *)
      best_cmd := (match patterns.(x).pat_action with
         | Move(z) -> Move(map_rot_dir d z)
         | z -> z)
    end
  done;
  !best_cmd
;;

(****** decision making ******)

let can_pickup robot pack =
  (robot.capacity - (package_weight pack) >= 0)

let make_move_from_route board robots r route =
  let (x,y,d) = Queue.peek route in
  let best_move = pattern_match_navigate board robots r d in
    { bid = 1; order = best_move }

let find_nearest_base board dists =
  let base = ref (1,1) in
  let best_dist = ref max_dist in
  for i = 1 to Array.length board - 1 do
    for j = 1 to Array.length board.(i) - 1 do
      if board.(i).(j).kind == BASE then
        let dist = dists.(i).(j) in
        if dist < !best_dist then
          base := (i,j);
          best_dist := dist
    done
  done;
  !base

(* Find nearest unexplored base or known package *)
let get_likely_package_location board dists unexplored_bases packages =
(* find_nearest_base board dists *)
  let base = ref (1,1) in
  let best_dist = ref max_dist in
  let check_dist x y =
    let dist = dists.(x).(y) in
(*  Printf.printf "Checking dist to %d %d -> %d\n" x y dist; *)
    if dist < !best_dist then (base := (x,y); best_dist := dist) in
  list_iter unexplored_bases (fun (x,y) -> check_dist x y);
  hash_iter packages (
    fun k v ->
      match v.pkg_loc with
        | Sitting(x,y) -> check_dist x y
        | Carried(r) -> ());
  if !best_dist = max_dist then raise Not_found
  else !base

let consider_dunk board dists robots packages r pkgs_here unexplored_bases =
  let one_side_water_one_side_open (x1,y1) (x2,y2) =
    let sq1 = board.(x1).(y1).kind in
    let sq2 = board.(x2).(y2).kind in
    match sq1,sq2 with
      | WATER,SPACE -> true
      | WATER,BASE -> true
      | SPACE,WATER -> true
      | BASE,WATER -> true
      | _ -> false in
  let dunkable robot =
    if board.(robot.x).(robot.y).is_beach then
      let dx = robot.x - r.x in
      let dy = robot.y - r.y in
      let farx = robot.x + dx in
      let fary = robot.y + dy in
      match board.(farx).(fary).kind with
        | WATER ->
            let side1x = robot.x + dy in
            let side1y = robot.y + dx in
            let side2x = robot.x - dy in
            let side2y = robot.y - dx in
            let nah = one_side_water_one_side_open (side1x,side1y) (side2x,side2y) in
            not nah
        | _ -> false
    else false in
  let robot_list = hash_values robots in
  let nearby_robots =
    list_filter robot_list
      (fun rob -> abs(rob.x - r.x) + abs(rob.y - r.y) = 1) in
  try
    let dunkee = list_find nearby_robots dunkable in
    let dir = deltas_to_dir (dunkee.x - r.x, dunkee.y - r.y) in
    Some(({ bid=1; order=Move(dir)}), "dunking")
  with Not_found -> None
  
let consider_drop board dists robots packages r pkgs_here unexplored_bases =
  let drop_list = List.fold_left
    (fun result item ->
        let p = Hashtbl.find packages item in
        if package_dest p = (r.x,r.y) then
          item::result
        else
          result
      ) [] r.packages in

  if drop_list != [] then 
    Some(({ bid=1; order=Drop(drop_list) }), "dropping")
  else
    None
;;

let thrashing r =
  if List.length r.history < max_history then false else

  let min_x = List.fold_left (fun a (x, y) -> min a x) r.x r.history
  and max_x = List.fold_left (fun a (x, y) -> max a x) r.x r.history
  and min_y = List.fold_left (fun a (x, y) -> min a y) r.y r.history
  and max_y = List.fold_left (fun a (x, y) -> max a y) r.y r.history in

  abs(max_x - min_x) + abs(max_y - min_y) <= 1
;;
    

(* if we're thrashing, stall for a turn *)
let consider_wait board dists robots packages r pkgs_here unexplored_bases =
  if Random.int 4 == 0 && thrashing r then
    Some({ bid=1; order=Pick([]) }, "stalling to try to unclog")
  else
    None
;;

(* check if we're standing on anything we can carry *)
let consider_pickup board dists robots packages r pkgs_here unexplored_bases =
  (* sum from list but don't go over max *)
  let sum_until list max =
    let rec sum_until' list max acc acc_list = match list with
      | [] -> (acc,acc_list)
      | hd::tl -> if (package_weight hd) + acc <= max
        then
          sum_until' tl max ((package_weight hd) + acc) (hd::acc_list)
        else
          (acc,acc_list) in
    sum_until' list max 0 [] in

  let weight_cmp p1 p2 = 
    (package_weight p2) - (package_weight p1) in

  let possible_pickups = list_filter pkgs_here (can_pickup r) in

  let total_pkgs = ref [] in
  let keep_going = ref true in
  let capacity_left = ref r.capacity in
  let src_dists = ref dists in

  if list_nonempty possible_pickups then (
    (* First, sort packages into groups by destination *)
    let (groups: (int*int, package list) Hashtbl.t ) = Hashtbl.create 10 in
    list_iter pkgs_here
      (fun p ->
         let dest = package_dest p in 
         if not (Hashtbl.mem groups dest) then Hashtbl.add groups dest [];
         let lst = Hashtbl.find groups dest in
         Hashtbl.replace groups dest (p::lst));
  
    (* Then sort that group by weight *)
    let (sorted_groups: (int*int, package list) Hashtbl.t ) = Hashtbl.create 10 in
    hash_iter groups
      (fun k v ->
         Hashtbl.add sorted_groups k (list_sort v weight_cmp));
      
    (* Find out how far each group's destination is *)
    let (distances: (int*int, int) Hashtbl.t ) = Hashtbl.create 10 in
    hash_iter groups
      (fun (x,y) v ->
         let dist = !src_dists.(x).(y) in
         Hashtbl.add distances (x,y) dist);
      
(*  Printf.fprintf !sock_log "******\n"; *)

    while !keep_going do
(*    Printf.fprintf !sock_log "\n";
      hash_iter groups
        (fun (x,y) v ->
           Printf.fprintf !sock_log
           "(%d,%d) %s\n" x y (String.concat " " (list_map v package_to_str)));
      Printf.fprintf !sock_log "***\n"; *)
      
      (* Then find out how much we can carry in each group, greedily *)
      let (carryable_weights: (int*int, (int * package list)) Hashtbl.t) = Hashtbl.create 10 in
      hash_iter sorted_groups
        (fun k v -> Hashtbl.add carryable_weights k (sum_until v !capacity_left));
    
(*    Printf.fprintf !sock_log "\n";
      hash_iter carryable_weights
        (fun (x,y) v ->
           Printf.fprintf !sock_log
           "(%d,%d) %d/%d {%s}\n" x y (fst v) (Hashtbl.find distances (x,y))
           (String.concat " " (list_map (snd v) package_to_str))); *)
      
      (* If there's more room, take the group with the most
         points/distance, where distance is from the first group we just
         took *)
      let best_ratio = ref 0.0 in
      let best_group = ref (-1,-1) in
      hash_iter carryable_weights
        (fun k (weight,lst) ->
           let ratio = float weight /. (float (Hashtbl.find distances k) +. 0.01) in
           if ratio > !best_ratio then (
             best_ratio := ratio;
             best_group := k));
      
      if !best_ratio > 0.0 then (
        let (x,y) = !best_group in
(*      Printf.fprintf !sock_log "Best group is (%d,%d)\n" x y; *)
        
        let (weight, pkgs_to_get_this_round) = Hashtbl.find carryable_weights !best_group in
        total_pkgs := pkgs_to_get_this_round @ !total_pkgs;
        capacity_left := !capacity_left - weight;
        
        keep_going := (!capacity_left > 0);

        if !keep_going then (
(*        Printf.fprintf !sock_log "Weight %d, remaining capacity %d\n" weight !capacity_left; *)

          (* Remove what we're taking from our hashes *)
          Hashtbl.replace groups !best_group
            (list_filter
               (Hashtbl.find groups !best_group)
               (fun pkg -> not (List.mem pkg pkgs_to_get_this_round)));
          Hashtbl.replace sorted_groups !best_group
            (list_sort (Hashtbl.find groups !best_group) weight_cmp);
          Hashtbl.replace carryable_weights !best_group (0,[]);
          src_dists := find_distances normal board x y
        )
      ) else
        keep_going := false;
    done;
    if list_nonempty !total_pkgs then
      Some ({bid=1; order=Pick( List.rev_map (fun item -> item.pkg_id) !total_pkgs )}, "pickup!")
    else None
  )
  else None
;;

(* if we're carrying any packages, deliver the nearest *)
let consider_deliver board dists robots packages r pkgs_here unexplored_bases =
  if r.packages != [] then begin
    let (best_d, choice) = 
      List.fold_left
        (fun (bestdist, best) item_num ->
          let item = Hashtbl.find packages item_num in
          (match item.pkg_dest with
             | Some(destx,desty) -> 
                 let thisdist = dists.(destx).(desty) in
                 if (thisdist < bestdist) then
                   (thisdist, item_num)
                 else
                   (bestdist, best)
             | None -> (bestdist, best)))
        (max_dist, List.hd r.packages)
        r.packages in

    (* find route to nearest *)

    let item = Hashtbl.find packages choice in

    let (route,reason) = match item.pkg_dest with
      | Some(x) -> find_good_route board dists robots r x
      | None -> (Queue.create(),"") in

    (* XXX if unreachable, they're all unreachable, so mark them as unreachable
       and drop them. TODO *)

    if Queue.is_empty route then begin
      Printf.printf " - unreachable package dest -";
      (* before we returned an arbitrary move; I think instead we should
         fallthrough to the next consider_X case *)
      None
    end else
      Some(make_move_from_route board robots r route, "going to dropoff-"^reason)
  end else
    None
;;

(* find nearest base *)
let consider_seekbase board dists robots packages r pkgs_here unexplored_bases =
    try 
      let (bx,by) = get_likely_package_location board dists unexplored_bases packages in
(*    Printf.printf "looking for route to %d %d\n" bx by; *)
      let (route,reason) = find_good_route board dists robots r (bx,by) in
      if Queue.is_empty route then (
        Printf.printf " - no bases reachable, nothing to do -";
        None
      ) else
        Some(make_move_from_route board robots r route, "going to base-"^reason)
    with
        Not_found -> None
;;

(* do something random *)
let consider_catchall board dists robots packages r pkgs_here unexplored_bases =
  Some(({bid=1; order=Pick([])}, "catchall"))
;;

 
let bidwar_range = 1;;

let major_defense r =
  max (r.bankroll/20) 10
;;

let minor_defense r =
  max (r.bankroll/100) 10
;;

let major_offense r =
  max (r.bankroll/100) 3
;;

(* return the amount we should bid in a war with nearby robots *)
let nearby_bidwar x y robots ignore default =
  Hashtbl.fold (fun k v res ->
      let d = abs(v.x-x) + abs(v.y-y) in
      if v != ignore && d <= bidwar_range && d>0 && v.bidwar > res then 
        v.bidwar 
      else 
        res)
    robots
    default
;;

(* Number of robots in the same location as r *)
let robots_here r robots =
  Hashtbl.fold (fun k v res -> 
    (if v.x = r.x && v.y = r.y then (res + 1) else res))
    robots
    0
;;

(* Number of robots at most d away from x y *)
let robots_delta_d x y d robots =
  Hashtbl.fold (fun k v res -> 
   (if abs(v.x-x) + abs(v.y-y) <= d then (res + 1) else res))
    robots
    0
;;

(* Is there a beach within dist of x y in the direction dir? *) 
let rec beach_nearby board dist dir x y =
  match dist with
  | 0 -> board.(x).(y).is_beach
  | v -> if board.(x).(y).is_beach then 
           true 
         else if board.(x).(y).kind == WALL then
           false
         else
           let (dx,dy) = dir_to_deltas dir in
           beach_nearby board (dist-1) dir (x+dx) (y+dy)
;;

(* can this robot push a thing on this spot into this type of terrain *)
let can_push_into board r x y k =
  let dx = x - r.x and dy = y - r.y in
    (* indexing into board should always be safe if dx+dy = 1 *)
    (* must be abs() of both!!! this was crashing *)
    (abs(dx) + abs(dy) = 1) && (board.(x+dx).(y+dy).kind = k)
;;

(* If there were a robot on this spot, could the passed-in robot dunk it? *)
let can_dunk_spot board r x y =
  can_push_into board r x y WATER
;;
      
(* If we're in danger of being dunked, that's a major problem *)
let bid_consider_avoid_dunk board my_robot robots dir sock =
  let can_dunk_us r = can_dunk_spot board r my_robot.x my_robot.y in
  if list_exists (hash_values robots) can_dunk_us then (
(*  Printf.fprintf sock.log "Special bid to avoid dunking\n"; *)
    Some(major_defense my_robot) )
  else
    None
;;

(* If we're moving to a beach and someone can dunk us after we move there,
   move slowly, so that if they predict we're going to move there, we don't
   arrive before they push us - this is a minor defense because it's not
   going to be a common offense (we hope) *)
(* bugfix: don't consider OURSELVES as candidates for this, which
   showed up on single-player boards *)
let bid_consider_avoid_predictive_dunk board my_robot robots dir sock =
  let (dx,dy) = dir_to_deltas dir in
  let can_dunk_us_later r = (r != my_robot && can_dunk_spot board r (my_robot.x+dx) (my_robot.y+dy)) in
  if list_exists (hash_values robots) can_dunk_us_later then (
(*  Printf.fprintf sock.log "Special bid to avoid predictive dunking\n"; *)
    Some(- minor_defense my_robot) )
  else 
    None
;;

(* If we can dunk someone by moving there, that's good offense *)
(* bugfix: there has to actually be a bot there! *)
let bid_consider_perform_dunk board my_robot robots dir sock =
  let (dx,dy) = dir_to_deltas dir in
  let at_target r = (r.x == my_robot.x+dx) && (r.y == my_robot.y+dy) in
  if can_dunk_spot board my_robot (my_robot.x+dx) (my_robot.y+dy)
  && list_exists (hash_values robots) at_target then
  (
(*  Printf.fprintf sock.log "Special bid to attempt dunking\n"; *)
    Some(major_offense my_robot) )
  else 
    None
;;

(* If we're carrying nothing and can only be pushed into a wall, whatever -
   this is for those width-two corridors *)
let bid_consider_wall_push board my_robot robots dir sock =
  let push_danger r = 
    let dx = abs(my_robot.x - r.x) and dy = abs(my_robot.y - r.y) in
(*  Printf.fprintf sock.log "%s is delta %d %d -> " r.id dx dy; *)
    let ok = (dx + dy = 1) && not (can_push_into board r my_robot.x my_robot.y WALL) in
(*  Printf.fprintf sock.log "%b\n" ok; *)
    ok
  in
  if not(any_nearby robots my_robot) ||
     my_robot.packages != [] || 
     list_exists (hash_values robots) push_danger then 
    None
  else (
(*  Printf.fprintf sock.log "Special bid to ignore wall-push\n"; *)
    Some(1) )
;;

(* If we're not carrying anything and the only pushes would move us where 
   we want to go, whatever *)
let bid_consider_helpful_push board my_robot robots dir sock =
  let unhelpful_push r = 
    let dx = my_robot.x - r.x and dy = my_robot.y - r.y and
        (dirx, diry) = dir_to_deltas dir in
      (abs(dx) + abs(dy) = 1) && not (dx == dirx && dy == diry)
  in
  if not(any_nearby robots my_robot) ||
     my_robot.packages != [] || 
     list_exists (hash_values robots) unhelpful_push then 
    None
  else (
(*  Printf.fprintf sock.log "Special bid to ignore helpful push\n"; *)
    Some(1) )
;;

(* Generic move evaluation *)
let bid_consider_generic_move board r robots dir sock =
  let (dx,dy) = dir_to_deltas dir in
    let bid1 = nearby_bidwar r.x r.y robots r 0 and
      bid2 = -(nearby_bidwar (r.x+dx) (r.y+dy) robots r 0)in
(*    Printf.fprintf sock.log "Move bid, bid1=%d bid2=%d\n" bid1 bid2; *)
      (* XXX if carrying packages then increment bid1 and decrement bid2 *)
      if bid1 != 0 && bid2 != 0 then
        Some(if abs(bid2) > abs(bid1) then -1 else 1)
      else if bid1 = 0 && bid2 = 0 then
        Some(1)
      else if bid1 = 0 then
        Some(bid2)
      else if bid2 = 0 then
        Some(bid1)
      else
        Some(1)
;;

(* Calculate bid due to moving *)
let calc_moving_bid board r robots dir sock =
  let funcs = [ bid_consider_avoid_dunk,"avoid dunk";
                bid_consider_avoid_predictive_dunk,"avoid pred dunk" ;
                bid_consider_perform_dunk,"perform dunk";
                bid_consider_wall_push,"wall push" ;
                bid_consider_helpful_push,"help push" ;
                bid_consider_generic_move,"generic move" ]
  and test (f,who) =
     try
     f board r robots dir sock
     with Invalid_argument(s) -> (Printf.printf "Invalid_argument(%s) in cmb %s" s who;None)
  in
  let res = iter_abort funcs test in
  match res with
    | Some(x) -> x
    | None -> 1 (* Hmm, none of the rules matched, weird *)
;;

let calc_bid board r robots command sock =
  match command with
  | No_order    -> 1
  | Pick( lst ) -> if List.length lst > 0 then 
                     let b = nearby_bidwar r.x r.y robots r 1 in
(*                     Printf.fprintf sock.log "Contested bidwar on Pick, bidding: %d\n" b; *)
                       b 
                   else 1
  | Drop( lst ) -> if List.length lst > 0 then 
                     let b = nearby_bidwar r.x r.y robots r 1 in
(*                     Printf.fprintf sock.log "Contested bidwar on Drop, bidding: %d\n" b; *)
                       b 
                   else 1
  | Move( dir ) -> let rh = (robots_here r robots) - 1 in
      (* if nobody's on our spot, do a normal bid: *)
      if rh = 0 then
        let b = calc_moving_bid board r robots dir sock in
(*        Printf.fprintf sock.log "Final move bid: %d\n" b; *)
          b
      (* otherwise, make sure to avoid getting pushed into/by the water: *)
      else if (any_nearby robots r) && (beach_nearby board rh dir r.x r.y) then 
        let b = - (minor_defense r) in
(*        Printf.fprintf sock.log "There are robots on our square and we're by the beach, so bidding %d\n" b; *)
          b
      (* otherwise whatever *)
      else
        1
;;

let max_bidwar my_robot = my_robot.bankroll / 100;;

let increase_bidwar my_robot robots last_bid =
  (* XXX this could actually consider which direction we were pushed from,
     and hence who might have pushed us. but for now it just assumes
     anyone nearby could have done so *)
  let f k v = (
    if (abs(v.x - my_robot.x) + abs(v.y - my_robot.y)) = 1 then
      let tmp_bw = min (max v.bidwar (last_bid + 2)) (max_bidwar my_robot) in
      let act_bw = min tmp_bw (v.bidwar + 2) in
      v.bidwar <- act_bw
  )
  in
    Hashtbl.iter f robots 
;;

let check_if_pushed updates robots my_robot cmd =
  try
    let my_first = List.find (fun x -> x.robot = my_robot) updates in
      match my_first.action with
      (* If our first action was a move, that's ok if we wanted to go that 
         way *)
      | UpdateMove(m) -> (
        match cmd.order with 
        | Move(dir) -> if dir != m then increase_bidwar my_robot robots cmd.bid
        | _ -> increase_bidwar my_robot robots cmd.bid)

      (* If our first action was a drop, that's ok if we wanted to drop stuff
         and every drop we did was on our list *)
      | UpdateDrop(d) -> (
        match cmd.order with
        | Drop(lst) -> (
          let f = fun update ->
            if update.robot = my_robot then
              match update.action with
              | UpdateDrop(d) ->
                if not (List.mem d lst) then true else false
	      | _ -> false
            else
              false
          in
            if List.exists f updates then increase_bidwar my_robot robots cmd.bid)
        | _-> increase_bidwar my_robot robots cmd.bid)

      (* If we didn't start out moving or dropping, we weren't pushed *)
      | _ -> ()
  with
    (* If we didn't make a move, we may have been pushed, or we may
       have purposely decided to do nothing *)
    Not_found -> 
      match cmd.order with
      | Pick(lst) -> if List.length lst > 0 then increase_bidwar my_robot robots cmd.bid
      | Drop(lst) -> if List.length lst > 0 then increase_bidwar my_robot robots cmd.bid
      | _ -> increase_bidwar my_robot robots cmd.bid
;;

let generate_command board robots packages r pkgs_here unexplored_bases sock =
  let dists = find_distances normal board r.x r.y in
  let funcs = [ consider_dunk,"dunk" ;
                consider_drop,"drop" ;
                consider_pickup,"pickup" ;
                consider_wait,"wait" ;
                consider_deliver,"deliver" ;
                consider_seekbase,"seekbase" ;
                consider_catchall,"catchall" ]
  and test (f, who) = f board dists robots packages r pkgs_here unexplored_bases in
  let res = iter_abort funcs test in
  match res with
    | Some(x,y) -> ({ bid = calc_bid board r robots x.order sock; order = x.order }, y)
    | None -> ({ bid=1; order=Pick([]); }, "fallthrough")
;;

(****** tracking info *******)

(* is x y the destination of pkg? *)
let at_destination pkg x y =
  match pkg.pkg_dest with
    | Some(dx,dy) -> dx = x && dy = y
    | None -> false

let definitely_not_at_destination pkg x y =
  match pkg.pkg_dest with
    | Some(dx,dy) -> dx <> x && dy <> y
    | None -> false

(* update list -> () *)
let process_updates robots updates packages my_robot cmd sock =
  let process_update update =
    let r = update.robot in
    match update.action with
      | UpdateMove(m) ->
          let (dx,dy) = dir_to_deltas m in
          r.x <- r.x + dx;
          r.y <- r.y + dy
      | UpdateX(x) -> r.x <- x
      | UpdateY(y) -> r.y <- y
      | UpdatePick(p) ->
          r.packages <- p :: r.packages;
          let pack = get_or_make_package p packages (Carried r) in
          r.capacity <- r.capacity - (package_weight pack);
          pack.pkg_loc <- Carried( r )
      | UpdateDrop(d) ->
          r.packages <- list_filter r.packages (fun p -> p != d);
          let pack = get_or_make_package d packages (Sitting (r.x,r.y)) in
          r.capacity <- r.capacity + (package_weight pack);
          pack.pkg_loc <- (Sitting (r.x,r.y) );
          let nearby_robots = robots_delta_d r.x r.y 1 robots in
(*        Printf.fprintf !sock_log "%d nearby robots\n" nearby_robots; *)
          let at_delivery = at_destination pack r.x r.y in
          let not_at_delivery = definitely_not_at_destination pack r.x r.y in
(*        Printf.fprintf !sock_log "at_delivery = %b\n" at_delivery; *)
(*        Printf.fprintf !sock_log "not_at_delivery = %b\n" not_at_delivery; *)
          if (not not_at_delivery) || nearby_robots <= 1 then (
            (* it was probably delivered *)
(*            Printf.fprintf !sock_log "%d was probably delivered\n" pack.pkg_id; *)
            Hashtbl.remove packages pack.pkg_id )
  in
  check_if_pushed updates robots my_robot cmd;
  list_iter updates process_update
;;

let clear_dead_robots robots =
  let dead_robots = ref [] in
  hash_iter robots (fun k v -> if not v.alive then dead_robots := k::!dead_robots);
  list_iter !dead_robots (fun k -> Hashtbl.remove robots k)
;;

(* list all package ids that packages thinks should be here at x y but
   are missing in pkgs_here *)
let compute_missing_packages packages pkgs_here x y =
  let missing_packages = ref [] in
  let ids_here = list_map pkgs_here (fun p -> p.pkg_id) in
  let check_package k v =
    match v.pkg_loc with
      | Carried(r) -> ()
      | Sitting(i,j) ->
          if x = i && y = j && not (List.mem v.pkg_id ids_here) then
            missing_packages := k :: !missing_packages in
  hash_iter packages check_package;
  !missing_packages

(* update our packages hash to reflect what's here at x y *)
let update_packages packages pkgs_here x y =
  (* update hash with new info *)
  list_iter pkgs_here (fun x -> Hashtbl.replace packages x.pkg_id x);
  (* If any packages that we expected to be here are not, then remove them *)
  let missing_packages = compute_missing_packages packages pkgs_here x y in
  list_iter missing_packages (fun p -> Hashtbl.remove packages p )


(* If this is a base, remove it from the unexplored list *)
let remove_from_unexplored_base_list board unexplored_bases (x,y) =
  match board.(x).(y).kind with
      BASE -> unexplored_bases := 
        list_filter !unexplored_bases (fun (i,j) -> i != x || j != y)
    | _ -> ()

let print_map board robot =
  for y = Array.length (board.(1)) - 2 downto 1 do
    for x = 1 to Array.length (board) - 2 do
      if y == robot.y && x == robot.x then
        print_char 'R'
      else if board.(x).(y).has_robots then
        print_char 'r'
      else if board.(x).(y).is_beach then
        print_char (square_ch_beach board.(x).(y).kind)
      else
        print_char (square_ch board.(x).(y).kind)
    done;
    print_newline()
  done 

let log_map board robots my_rob f =
  let width  = Array.length board
  and height = Array.length board.(1) in
  let r = Array.init width (fun i -> Array.make height ' ') in
  Hashtbl.iter (fun k v -> r.(v.x).(v.y) <- v.name) robots;
  for y = height-2 downto 1 do
    for x = 1 to width-2 do
      if x == my_rob.x && y == my_rob.y then
        output_char f 'R'
      else if r.(x).(y) != ' ' then
        output_char f r.(x).(y)
      else if board.(x).(y).is_beach then
        output_char f (square_ch_beach board.(x).(y).kind)
      else
        output_char f (square_ch board.(x).(y).kind)
    done;
    output_char f '\n'
  done;
  Hashtbl.iter (fun k v ->
       if v != my_rob then Printf.fprintf f "%s:%c\n" v.id v.name) robots
;;

let mark_robot_positions board robots r =
  Hashtbl.iter (fun k v ->
    if abs(r.x - v.x) + abs(r.y - v.y) <= avoid_range then
       board.(v.x).(v.y).has_robots <- true)
    robots;;

let unmark_robot_positions board robots r =
  Hashtbl.iter (fun k v -> board.(v.x).(v.y).has_robots <- false) robots;;

(***********************)
let take_a_turn
    (board: square array array)
    (robots: (string, robot) Hashtbl.t)
    (packages: (int, package) Hashtbl.t)
    (unexplored_bases: (int*int) list ref)
    (r: robot)
    (sock: io_interface)
    (dbg_cmd: bool) =

  let (pkgs_here:package list) = get_package_info r.x r.y sock in
  update_packages packages pkgs_here r.x r.y;

  remove_from_unexplored_base_list board unexplored_bases (r.x,r.y);

  mark_robot_positions board robots r;
 
  board.(r.x).(r.y).has_robots <- false;
  let (cmd,res) = 
    generate_command board robots packages r pkgs_here !unexplored_bases sock
  in

  unmark_robot_positions board robots r;

    if dbg_cmd then print_endline("cmd: " ^ order_to_str(cmd.order));
(*   Printf.fprintf sock.log "@%2d,%2d: %s (bid=%d bankroll=%d)\n" r.x r.y res cmd.bid r.bankroll; *)
    Printf.printf           "@%2d,%2d: %s (bid=%d bankroll=%d)\n" r.x r.y res cmd.bid r.bankroll;

  ignore (send_command cmd sock);

  hash_iter robots (fun k v -> (v.alive <- false));
  hash_iter robots (fun k v ->
    if List.length v.history >= max_history 
      then v.history <- (v.x, v.y) :: (List.rev (List.tl (List.rev v.history)))
      else v.history <- (v.x, v.y) :: v.history
  );

  let updates = receive_updates robots sock in
  process_updates robots updates packages r cmd sock;

  r.bankroll <- r.bankroll - abs(cmd.bid);

  clear_dead_robots robots

let print_packages packages =
  hash_iter packages (fun k v -> print_endline (package_to_str v))

(****** Startup ******)

let main host port dbg_options = 
  let dbg_display = List.mem "map" dbg_options in
  let dbg_packages = List.mem "packages" dbg_options in 
  let dbg_command = List.mem "commands" dbg_options in
  let dbg_bid = List.mem "bid" dbg_options in
  let dbg_predict = List.mem "predict" dbg_options in
  print_endline "[connecting]"; flush stdout;
  let sock = connect host port in
  sock_log := sock.log;
  print_endline "[connected]"; flush stdout;
  ignore (send_ln sock "Player") ;
  flush stdout;
  let board, unexplored_bases = read_board sock in
  let (myid, capacity, bankroll) = read_my_robot sock in
  let robots = read_all_robots sock in       (* Hashtbl int -> robot *)
  let packages = Hashtbl.create 100 in
  let my_robot = merge_robot robots (myid, capacity, bankroll) in
  for turns = 1 to bankroll do
    flush stdout;

(*  if turns mod 10 == 1 then log_map board robots my_robot sock.log; *)

    if dbg_predict then make_predictions stdout robots;
    if dbg_predict then make_predictions sock.log robots;

    take_a_turn board robots packages unexplored_bases my_robot sock dbg_command;

    (if dbg_display then 
       (print_map board my_robot;
        ignore (read_line())));

    (if dbg_packages then 
       (print_packages packages;
        ignore (read_line())));

    flush stdout
  done;;

let get_options argv =
  let len = Array.length argv in
  if len <= 3 then []
  else Array.to_list (Array.sub argv 3 (len-3) )
;;    

print_endline "Hello!";;
let options = get_options Sys.argv;;
main Sys.argv.(1) (int_of_string Sys.argv.(2)) options
